import { View, PureCollection } from '@datx/core';
import { Response as BaseResponse, IResponseObject } from '@datx/network';
import { IJsonapiModel } from './interfaces/IJsonapiModel';
import { IJsonApiObject, ILink } from './interfaces/JsonApi';
export declare class NetworkResponse<T extends IJsonapiModel> extends BaseResponse<T> {
    /**
     * API response metadata
     *
     * @type {object}
     * @memberOf Response
     */
    get meta(): object | undefined;
    /**
     * API response links
     *
     * @type {object}
     * @memberOf Response
     */
    get links(): Record<string, ILink> | undefined;
    /**
     * The JSON API object returned by the server
     *
     * @type {JsonApi.IJsonApiObject}
     * @memberOf Response
     */
    get jsonapi(): IJsonApiObject | undefined;
    /**
     * First data page
     *
     * @type {Promise<Response>}
     * @memberOf Response
     */
    first?: () => Promise<NetworkResponse<T>>;
    /**
     * Previous data page
     *
     * @type {Promise<Response>}
     * @memberOf Response
     */
    prev?: () => Promise<NetworkResponse<T>>;
    /**
     * Next data page
     *
     * @type {Promise<Response>}
     * @memberOf Response
     */
    next?: () => Promise<NetworkResponse<T>>;
    /**
     * Last data page
     *
     * @type {Promise<Response>}
     * @memberOf Response
     */
    last?: () => Promise<NetworkResponse<T>>;
    get views(): Array<View>;
    constructor(response: IResponseObject, collection?: PureCollection, overrideData?: T, views?: Array<View>);
    /**
     * Function called when a link is being fetched. The returned value is cached
     *
     * @private
     * @param {string} name Link name
     * @returns Promise that resolves with a Response object
     *
     * @memberOf NetworkResponse
     */
    private __fetchLink;
}
